<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <link rel="icon" href="./assets/favicon.ico" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <meta name="theme-color" content="#000000" />

    <title>Locust</title>
    <script type="module" crossorigin src="./assets/index-84c63e70.js"></script>
  </head>
  <body>
    <div id="root"></div>

    <script>
      window.templateArgs = {{ template_args|tojson }}
    </script>
    
  </body>
</html>
